/* 
 * ROPStop is Copyright (c) 2012-2014 Barton P. Miller. Written by Emily R. Jacobson.
 *
 * This file is part of ROPStop.
 *
 * ROPStop is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * ROPStop is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with ROPStop.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>

#include "Instruction.h"
#include "InstructionDecoder.h"

#include "CodeObject.h"
#include "globals.h"
#include "walker.h"
#include "CodeSource.h"
#include "procstate.h"
#include "validRA.h"

using namespace std;
using namespace Dyninst;

std::map<ParseAPI::CodeObject*, std::map<Address, ParseAPI::Function*> > preproc;
std::tr1::unordered_map<Address, std::pair<std::string, Address> > cachedLibs;
std::tr1::unordered_map<Address, ParseAPI::Function*> framefuncs;

bool isVsyscallPage(std::string libName)
{
    size_t str = libName.find("vsyscall");
    if (str != string::npos) {
        return true;
    } else {
        return false;
    }
}

Address getRelAddr(Address addr, std::string libName, Address libLoad)
{
    Address relAddr = addr - libLoad;

    if (isVsyscallPage(libName)) {
        Address magic_constant = 0xffffe000;
        relAddr = relAddr + magic_constant;
    }

    return relAddr;
}

ParseAPI::SymtabCodeSource* getCodeSource(Stackwalker::Walker * walker, string name, Address load, unsigned size)
{
    // Check in our cache for this code source
    std::tr1::unordered_map<string, ParseAPI::SymtabCodeSource*>::iterator found = srcs.find(name);
    if (found != srcs.end()) {
        return found->second;
    }

    SymtabAPI::Symtab * st;
    // Try opening the regular file
    if (!SymtabAPI::Symtab::openFile(st, name)) {
        if (DEBUG) fprintf(stderr, "openFile(%s) failed: %s\n",
                name.c_str(),
                SymtabAPI::Symtab::printError(SymtabAPI::Symtab::getLastSymtabError()).c_str());
        
        if (!walker) {
            return NULL;
        }

        // Try opening a memory image instead
        if (size < 0) {
            return NULL;
        }
       
        if (DEBUG) fprintf(stderr, "getCodeSource: calling openFile with memory image instead\n");
        char * mem_img = (char*)malloc(size);
        assert(mem_img);
        if (!walker->getProcessState()->readMem(mem_img, load, size)) {
            return false;
        }

        if (!SymtabAPI::Symtab::openFile(st, mem_img, size, name)) {
            if (DEBUG) fprintf(stderr, "openFile(%s)[mem_img version] failed: %s\n",
                    name.c_str(),
                    SymtabAPI::Symtab::printError(SymtabAPI::Symtab::getLastSymtabError()).c_str());
            return NULL;
        }
    }

    // Store in cache and return
    ParseAPI::SymtabCodeSource * cs = new ParseAPI::SymtabCodeSource(st);
    srcs.insert(make_pair(name, cs));

    return cs;
}

ParseAPI::CodeObject * getCodeObject(Stackwalker::Walker * walker, string name, Address load, unsigned size)
{
    // Check in our cache for this code object
    std::tr1::unordered_map<string,ParseAPI::CodeObject*>::iterator found = objs.find(name);
    if (found != objs.end()) {
        return found->second;
    }

    // Get the corresponding code source
    ParseAPI::CodeSource * cs = getCodeSource(walker, name, load, size);
    if (!cs) {
        return NULL;
    }

    // Store in cache
    ParseAPI::CodeObject * co = new ParseAPI::CodeObject(cs);
    objs.insert(make_pair(name, co));

    // Parse the code object (and gap parse)
    fprintf(stderr, "Forced parse of %s\n", name.c_str());
    co->parse();
    if (!isVsyscallPage(name)) {
      fprintf(stderr, "\t and parsing gaps\n");
      vector<ParseAPI::CodeRegion*> const & regions = cs->regions();
      vector<ParseAPI::CodeRegion*>::const_iterator rIter;
      for (rIter = regions.begin(); rIter != regions.end(); ++rIter) {
            // We only want to gap parse regions that are code!!
            if ((*rIter)->isCode((*rIter)->low())) {
                co->parseGaps(*rIter);
            } else {
                if (DEBUG) fprintf(stderr, "\t\t skipping region at 0x%lx\n", (*rIter)->low());
            }
      }
    }
    fprintf(stderr, "\t Done parsing\n");
    // Pre-process stuff...
    std::map<Address, ParseAPI::Function*> preprocLocal;

    for (ParseAPI::CodeObject::funclist::iterator iter1 = co->funcs().begin();
            iter1 != co->funcs().end(); ++iter1) {
        for (std::vector<ParseAPI::FuncExtent*>::const_iterator iter2 = (*iter1)->extents().begin(); 
                iter2 != (*iter1)->extents().end(); ++iter2) {
            preprocLocal[(*iter2)->end()] = *iter1;
        }   
    }
    
    preproc.insert(make_pair(co, preprocLocal));

    return co;
}

std::set<ParseAPI::CodeRegion *> getCodeRegions(ParseAPI::CodeSource * cs, Address addr)
{
    std::set<ParseAPI::CodeRegion*> regions;

    if (!cs->findRegions(addr, regions)) {
        if (DEBUG) fprintf(stderr, "\t\t findRegions failed; could not find 0x%lx\n", 
                addr);
    }

    return regions;
}

bool getLibraryAndSize(Stackwalker::Walker * walker, Address pc, std::string & name, Address & load, unsigned & size) 
{
    std::tr1::unordered_map<Address, std::pair<std::string, Address> >::iterator cIter;
    cIter = cachedLibs.find(pc);
    if (cIter != cachedLibs.end()) {
        name = cIter->second.first;
        load = cIter->second.second;
        return true;
    }

    Stackwalker::LibraryState * libState = walker->getProcessState()->getLibraryTracker();
    if (!libState) {
        return false;
    }

    Stackwalker::LibAddrPair lib;
    if (!libState->getLibraryAndSizeAtAddr(pc, lib, size)) {
        return false;
    }

    name = lib.first;
    load = lib.second;
    
    cachedLibs.insert(make_pair(pc, make_pair(name,load)));

    return true;
}

ParseAPI::Function * getFrameFunc(Stackwalker::Walker * walker, Address insn)
{
    /* Check if we have something cached */
    std::tr1::unordered_map<Address, ParseAPI::Function*>::iterator found = framefuncs.find(insn);
    if (found != framefuncs.end()) {
        return found->second;
    }

    std::string libName;
    Address libLoad;
    unsigned libSize;
    if (!getLibraryAndSize(walker, insn, libName, libLoad, libSize)) {
        return NULL;
    }

    ParseAPI::SymtabCodeSource * cs = getCodeSource(walker, libName, libLoad, libSize);
    if (!cs) return NULL;
    ParseAPI::CodeObject * co = getCodeObject(walker, libName, libLoad, libSize);

    Address relAddr = getRelAddr(insn, libName, libLoad);
    std::set<ParseAPI::CodeRegion *> regions = getCodeRegions(cs, relAddr);

    std::set<ParseAPI::Function*> funcs;
    ParseAPI::CodeRegion * cr;
    std::set<ParseAPI::CodeRegion*>::iterator rIter;
    for (rIter = regions.begin(); rIter != regions.end(); ++rIter) {
        cr = *rIter;
        if (funcs.size() > 1) {
            fprintf(stderr, "WARNING: getFrameFunc returning last func in set (multiple available)\n");
        }
        if (co->findFuncs(cr, relAddr, funcs)) {

            ParseAPI::Function* func = *(funcs.rbegin());
            // Store the cache!
            framefuncs.insert(make_pair<Address, ParseAPI::Function*>(insn, func));

            return (func);
        }
    }
    return NULL;
}

// THIS SHOULD ONLY BE USED ON THE VSYSCALL PAGE; IT CAN RETURN THE WRONG CALL INSTRUCTION OTHERWISE.
bool getPrevInsn(Stackwalker::Walker * walker,
        ProcControlAPI::Process::const_ptr proc,
        Address addr,
        InstructionAPI::Instruction::Ptr & insn,
        Address & insn_addr)
{
    int max_instr_size = 16;

    Stackwalker::ProcessState * procState = walker->getProcessState();
    Address call;

    unsigned char buffer[max_instr_size];
    for (unsigned size = max_instr_size; size > 0; size--) {
        call = addr - size;
        if (procState->readMem(buffer, call, size)) break;
    }

    unsigned char * bufferPtr = buffer;

    InstructionAPI::Instruction::Ptr tmp;
    InstructionAPI::Instruction::Ptr prevInsn;
    unsigned int aligned;
    int insn_size;
    for (unsigned size = addr - call; size > 0; size--) {

        InstructionAPI::InstructionDecoder d(bufferPtr, size, arch);
        
        tmp = d.decode();
        prevInsn = tmp;

        aligned = 0;

        // Decode all instructions in the buffer
        while (tmp) {
            if (tmp->size() > size) break;
           
            if (isInsnACall(tmp)) {
                fprintf(stderr, "\t\t\t getPrevInsn: found %s\n", tmp->format().c_str());
                insn = tmp;
                insn_addr = call + aligned;
                return true;
            }

            aligned += tmp->size();
            prevInsn = tmp;
            tmp = d.decode();
        }

        // prevInsn was the last valid instruction we found
        // is it (a) aligned
        if ( (aligned == size) && (isInsnACall(prevInsn))) {
            insn = prevInsn;
            insn_size = insn->size();
            insn_addr = addr - insn_size;
            return true;
        } else {
           fprintf(stderr, "\t\t\t getPrevInsn: %s not a call\n", prevInsn->format().c_str()); 
                
            bufferPtr++;
            continue;
        }
    }

    return false;
}


