/* 
 * ROPStop is Copyright (c) 2012-2014 Barton P. Miller. Written by Emily R. Jacobson.
 *
 * This file is part of ROPStop.
 *
 * ROPStop is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * ROPStop is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with ROPStop.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <asm/unistd.h>

#include "PCProcess.h"
#include "Event.h"
#include "PCErrors.h"

#include "Symtab.h"

#include "CodeObject.h"
#include "CodeSource.h"
#include "callbacks.h"
#include "globals.h"
#include "accessors.h"

using namespace std;
using namespace Dyninst;

void createCallback(ProcControlAPI::Process::ptr proc,
        ParseAPI::Function * func,
        Address addr);
std::set<int> monitoredSyscalls;
int COUNTER = 0;


ProcControlAPI::Process::ptr NEW_PROCESS = ProcControlAPI::Process::ptr();
ProcControlAPI::Process::ptr NEW_EXEC = ProcControlAPI::Process::ptr();
ProcControlAPI::Thread::ptr NEW_THREAD = ProcControlAPI::Thread::ptr();

void printSyscall(int syscall, Address addr) {
    COUNTER++;        
}

ProcControlAPI::Process::cb_ret_t on_event(ProcControlAPI::Event::const_ptr event)
{
    ProcControlAPI::EventType type = event->getEventType();
    ProcControlAPI::Process::const_ptr proc = event->getProcess();
    std::string name = type.name();
    if (DEBUG) fprintf(stderr, "%d: Event %s occurred: ", proc->getPid(), name.c_str());

    ProcControlAPI::Process::cb_ret_t ret = ProcControlAPI::Process::cbProcContinue; /* default */

    switch (type.code()) {
        case ProcControlAPI::EventType::Syscall:
            {
                ProcControlAPI::EventSyscall::const_ptr syscallEvent = event->getEventSyscall();
                Address addr = syscallEvent->getAddress();
                int syscall = syscallEvent->getSyscallNumber();
                if (ALL_SYSCALLS) {
                    /* Stop the process to monitor all system calls */
                    printSyscall(syscall, addr);
                    ret = ProcControlAPI::Process::cbProcStop;
                } else {
                    /* Stop the process to monitor select system calls */
                    if (monitoredSyscalls.find(syscall) != monitoredSyscalls.end()) {
                        printSyscall(syscall, addr);
                        ret = ProcControlAPI::Process::cbProcStop;
                    }
                    else {
                    }
                }
                break;
            }
        case ProcControlAPI::EventType::Fork: 
            {
                ProcControlAPI::EventFork::const_ptr forkEvent = event->getEventFork();
                ProcControlAPI::Process::const_ptr childProc = forkEvent->getChildProcess();
                PID childPID = childProc->getPid();
                if (DEBUG) fprintf(stderr, "\t new process id = %d\n", childPID);
                ret  = ProcControlAPI::Process::cb_ret_t(ProcControlAPI::Process::cbProcContinue, 
                        ProcControlAPI::Process::cbProcStop);
                NEW_PROCESS = pc_const_cast<ProcControlAPI::Process>(childProc);
                break;
            }
        case ProcControlAPI::EventType::ThreadCreate:
            {
                ProcControlAPI::EventNewThread::const_ptr threadEvent = event->getEventNewThread();
                ProcControlAPI::Thread::const_ptr newThread = threadEvent->getNewThread();
                LWP threadPID = newThread->getLWP();
                if (DEBUG) fprintf(stderr, "\t new thread id = %d\n", threadPID);
                ret  = ProcControlAPI::Process::cb_ret_t(ProcControlAPI::Process::cbThreadContinue, 
                        ProcControlAPI::Process::cbThreadStop);
                break;
            }
        case ProcControlAPI::EventType::Exec: 
            {
                ProcControlAPI::EventExec::const_ptr execEvent = event->getEventExec();
                std::string execPath = execEvent->getExecPath();
                if (DEBUG) fprintf(stderr, "\t new exe path = %s\n", execPath.c_str());
                ParseAPI::CodeObject* co = getCodeObject(NULL, execPath, 0, 0);
                if (!co) {
                    fprintf(stderr, "\t\t getCodeObject failed\n");
                }
                ret  = ProcControlAPI::Process::cb_ret_t(ProcControlAPI::Process::cbProcContinue, 
                        ProcControlAPI::Process::cbProcStop);
                NEW_EXEC = pc_const_cast<ProcControlAPI::Process>(proc); 

                break;
            }
        case ProcControlAPI::EventType::Library: 
            {
                ProcControlAPI::EventLibrary::const_ptr libraryEvent = event->getEventLibrary();
                const std::set<ProcControlAPI::Library::ptr> & libs = libraryEvent->libsAdded();
                if (libs.size() == 0) {
                    if (DEBUG) fprintf(stderr, "\n");
                    return ProcControlAPI::Process::cbProcContinue;
                }

                std::set<ProcControlAPI::Library::ptr>::const_iterator lIter;
                for (lIter = libs.begin(); lIter != libs.end(); ++lIter) {
                    ProcControlAPI::Library::ptr library = *lIter;
                    if (DEBUG) fprintf(stderr, "\t added library \"%s\" @ 0x%lx", 
                            library->getName().c_str(), 
                            library->getLoadAddress());
                    ParseAPI::CodeObject* co = getCodeObject(NULL, library->getName(), 0, 0);
                    if (!co) {
                        fprintf(stderr, "...getCodeObject failed");
                    }
                }
                if (DEBUG) fprintf(stderr, "\n");
                
                break;
            }
        default:
            {
                if (DEBUG) fprintf(stderr, "\t No handler defined for this event type\n");
                break;

            }
    }  

    return ret;

}

void addCallbacks(ProcControlAPI::Process::ptr proc)
{
    fprintf(stderr, "adding callbacks\n");

    ProcControlAPI::Process::registerEventCallback(ProcControlAPI::EventType(ProcControlAPI::EventType::Pre, 
                ProcControlAPI::EventType::Syscall), 
            on_event);
    ProcControlAPI::Process::registerEventCallback(ProcControlAPI::EventType::Fork, on_event);
    ProcControlAPI::Process::registerEventCallback(ProcControlAPI::EventType::UserThreadCreate, on_event);
    ProcControlAPI::Process::registerEventCallback(ProcControlAPI::EventType::Exec, on_event);
    ProcControlAPI::Process::registerEventCallback(ProcControlAPI::EventType::Library, on_event);
}
