/* 
 * ROPStop is Copyright (c) 2012-2014 Barton P. Miller. Written by Emily R. Jacobson.
 *
 * This file is part of ROPStop.
 *
 * ROPStop is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * ROPStop is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with ROPStop.  If not, see <http://www.gnu.org/licenses/>.
 */

/* Includes */
#include <stdio.h>
#include <getopt.h>
#include <iostream>
#include <sstream>
#include <assert.h>

#include "dyntypes.h"

/* ProcControl includes */
#include "PCProcess.h"
#include "PCErrors.h"

/* Stackwalker includes */
#include "walker.h"
#include "frame.h"
#include "basetypes.h"
#include "framestepper.h"
#include "steppergroup.h"
#include "swk_errors.h"

/* InstructionAPI includes */
#include "Instruction.h"
#include "InstructionDecoder.h"

/* ParseAPI includes */
#include "CFG.h"
#include "CodeObject.h"

/* SymtabAPI includes */
#include "Symtab.h"
#include "Function.h"

using namespace std;
using namespace Dyninst;

#include "manualWalk.h"
#include "validRA.h"
#include "printers.h"
#include "globals.h"
#include "accessors.h"

std::tr1::unordered_map<Address, ParseAPI::Function*> cachedCallees;

// Much of this logic is from dyninstAPI/src/unix.C, block_instance::findCallee() and its subsidiary functions.

bool findFuncByEntry(Stackwalker::Walker * walker,
        Address targetAddr,
        ParseAPI::Function *&func)
{
    if (DEBUG) fprintf(stderr, "\t\t\t findFuncByEntry: looking for function at target 0x%lx\n", targetAddr);
    ParseAPI::SymtabCodeSource * cs = NULL;
    ParseAPI::CodeObject * co = NULL;

    std::string libName;
    Address libLoad;
    unsigned libSize;
    if (!getLibraryAndSize(walker, targetAddr, libName, libLoad, libSize)) {
        if (DEBUG) fprintf(stderr, "\t\t\t findFuncByEntry: getLibraryAndSize failed\n");
        return false;
    } else {
        if (DEBUG) fprintf(stderr, "\t\t\t findFuncByEntry: library \"%s\" loaded at 0x%lx\n", libName.c_str(), libLoad);

        cs = getCodeSource(walker, libName, libLoad, libSize);
        if (!cs) {
            if (DEBUG) fprintf(stderr, "\t\t findFuncByEntry: getCodeSource failed\n");
            return false;
        }

        co = getCodeObject(walker, libName, libLoad, libSize);                        
        if (!co) {
            if (DEBUG) fprintf(stderr, "\t\t\t findFuncByEntry: getCodeObject failed\n");
            return false;
        }
        if (DEBUG) fprintf(stderr, "\t\t\t findFuncByEntry: got CodeObject\n");
    }

    // Get CodeRegion
    Address relAddr = getRelAddr(targetAddr, libName, libLoad);
    std::set<ParseAPI::CodeRegion *> regions = getCodeRegions(cs, relAddr);

    if (regions.size() == 0) {
        return false;
    }

    ParseAPI::CodeRegion * cr;
    std::set<ParseAPI::CodeRegion *>::iterator rIter;
    for (rIter = regions.begin(); rIter != regions.end(); ++rIter) {
        cr = *rIter;

        // What function(s) is associated with the target of the edge?
        func = co->findFuncByEntry(cr, targetAddr-libLoad);
        if (!func) {
            if (DEBUG) fprintf(stderr, "\t\t\t findFuncByEntry: no function exists at this entry point\n");
            continue;
        } else {
            if (DEBUG) fprintf(stderr, "\t\t\t findFuncByEntry: target func = %s\n", func->name().c_str());
            // Add result to our cache for use later
            return true;
        }
    } 
    return false;
}

// modeled after dynPCProcess-x86.C, PCProcess::hasBeenBound
bool hasBeenBound(Stackwalker::Walker * walker, 
        const SymtabAPI::relocationEntry & entry,
        ParseAPI::Function *& func,
        Address base_addr)
{
    Address got_entry = entry.rel_addr() + base_addr;
    Address bound_addr = 0;

    unsigned tmp = 0;

    // FIXME
    if (!walker->getProcessState()->readMem(&tmp, got_entry, sizeof(unsigned))) {
        if (DEBUG) fprintf(stderr, "\t\t\t\t hasBeenBound: readMem failed: %s\n", ProcControlAPI::getLastErrorMsg());
        return false;
    }
    bound_addr = (Address) tmp;
    
    if (! (bound_addr == (entry.target_addr()+6+base_addr))) {
        // the callee function has been bound by the runtime linker
        // find function and return
        if (!findFuncByEntry(walker, bound_addr, func)) {
            if (DEBUG) fprintf(stderr, "hasBeenBound: findFuncByEntry failed\n");
            return false;
        } else {
            return true;
        }
    }

    return false;
}

std::vector<ParseAPI::Function *> getEntryFunc(ParseAPI::Block * block)
{
    std::vector<ParseAPI::Function*> funcs;
    block->getFuncs(funcs);

    return funcs;
}

std::vector<ParseAPI::Function *> getTailCallee(ParseAPI::Edge * edge)
{

    ParseAPI::Block * trgBlock = edge->trg();
    return getEntryFunc(trgBlock);
}

std::vector<ParseAPI::Function *> getCallee(ParseAPI::Edge * edge)
{
    ParseAPI::Block * trgblock = edge->trg();
    return getEntryFunc(trgblock);
}

bool findTailCallee(Stackwalker::Walker * walker,
        ParseAPI::Block * srcblock,
        ParseAPI::Edge * edge,
        ParseAPI::SymtabCodeSource * srcCS,
        ParseAPI::Function *&func,
        Address libLoad)
{
    // Check if we have a cached result
    std::tr1::unordered_map<Address,ParseAPI::Function*>::iterator cacheIter;
    cacheIter = cachedCallees.find(edge->src()->lastInsnAddr());
    if (cacheIter != cachedCallees.end()) {
        func = cacheIter->second; 
        if (DEBUG) fprintf(stderr, "\t\t findTailCallee: found cached callee function = %s\n", func->name().c_str());
        return true;
    }
    
    // Special case; found sink edge
    if (edge->sinkEdge()) {
        if (DEBUG) fprintf(stderr, "\t\t\t findCallee: call edge has unresovled control flow target; conservatively assume edge exists to callee\n");
        return true;
    }

    ParseAPI::Block * target = edge->trg();
    Address targetAddr = target->start();

    // Get the relocation information for this image
    SymtabAPI::Symtab * symtab = srcCS->getSymtabObject();
    std::vector<SymtabAPI::relocationEntry> fbt;
    if (!symtab->getFuncBindingTable(fbt)) {
        if (DEBUG) fprintf(stderr, "\t\t\t findTailCallee: no relocation information for this Symtab\n");
        return false;
    }

    std::map<Address, std::string> pltFuncs = srcCS->linkage();

    // find the target address in the list of relocation entries
    if (DEBUG) fprintf(stderr, "\t\t\t findTailCallee: targetAddr = 0x%lx\n", targetAddr);
    if (pltFuncs.find(targetAddr) == pltFuncs.end()) {
        if (DEBUG) fprintf(stderr, "\t\t\t findTailCallee: targetAddr not found in pltFuncs\n");
        // Check if we know this already
        func = getTailCallee(edge)[0];
        if (func) {
            if (DEBUG) fprintf(stderr, "\t\t\t findTailCallee: found callee using getTailCallee() = %s\n", func->name().c_str());
            return true;
        }
    } else {
        for (u_int i = 0; i < fbt.size(); i++) {
            if (fbt[i].target_addr() == targetAddr) {
                // check to see if this function has been bound yet
                if (hasBeenBound(walker, fbt[i], func, libLoad)) {
                    cachedCallees[edge->src()->lastInsnAddr()] = func;
                    return true;
                }
            }
        }

        const char * target_name = pltFuncs[targetAddr].c_str();

        std::vector<SymtabAPI::Function *> pdfv;
        if (symtab->findFunctionsByName(pdfv, target_name, SymtabAPI::mangledName)) {
            SymtabAPI::Function * sfunc = pdfv[0];

            Offset funcOffset = sfunc->getOffset();
            if (!findFuncByEntry(walker, funcOffset, func)) {
                if (DEBUG) fprintf(stderr, "\t\t\t findTailCallee: findFuncByOffset failed\n");
                return false;
            }

            cachedCallees[edge->src()->lastInsnAddr()] = func;
            return true;
        }
    }

    if (DEBUG) fprintf(stderr, "\t\t\t findTailCallee: reached end of function, returning false\n");
    return false;

}

bool findCalleeFunc(Stackwalker::Walker * walker,
        Address callInsn,
        ParseAPI::SymtabCodeSource * srcCS,
        ParseAPI::Function *&func,
        Address libLoad,
        Address targetAddr)
{
    // Get the relocation information for this image
    SymtabAPI::Symtab * symtab = srcCS->getSymtabObject();
    std::vector<SymtabAPI::relocationEntry> fbt;
    if (!symtab->getFuncBindingTable(fbt)) {
        if (DEBUG) fprintf(stderr, "\t\t\t findCallee: no relocation information for this Symtab\n");
        return false;
    }

    std::map<Address, std::string> pltFuncs = srcCS->linkage();

    // find the target address in the list of relocation entries
    if (DEBUG) fprintf(stderr, "\t\t\t findCallee: targetAddr = 0x%lx, libLoad = 0x%lx\n", targetAddr, libLoad);
    if (pltFuncs.find(targetAddr) == pltFuncs.end()) {
        if (DEBUG) fprintf(stderr, "\t\t\t findCallee: targetAddr not found in pltFuncs\n");
        // Check if we know this already
        if (findFuncByEntry(walker, targetAddr+libLoad, func)) {
            if (DEBUG) fprintf(stderr, "\t\t\t findCalleeFunc: found callee using findFuncByEntry() = %s\n", func->name().c_str());
            return true;
        }
    } else {
        for (u_int i = 0; i < fbt.size(); i++) {
            if (fbt[i].target_addr() == targetAddr) {
                // check to see if this function has been bound yet
               if (hasBeenBound(walker, fbt[i], func, libLoad)) {
                    cachedCallees[callInsn] = func;
                    return true;
                }
            }
        }

        const char * target_name = pltFuncs[targetAddr].c_str();

        std::vector<SymtabAPI::Function *> pdfv;
        if (symtab->findFunctionsByName(pdfv, target_name, SymtabAPI::mangledName)) {
            SymtabAPI::Function * sfunc = pdfv[0];

            Offset funcOffset = sfunc->getOffset();
            
            if (!findFuncByEntry(walker, funcOffset, func)) {
                if (DEBUG) fprintf(stderr, "\t\t\t findCallee: findFuncByOffset failed\n");
                return false;
            }

            cachedCallees[callInsn] = func;
            return true;
        }
    }

    return false;
}

bool findCallee(Stackwalker::Walker * walker,
        ParseAPI::Block * srcblock,
        ParseAPI::SymtabCodeSource * srcCS,
        Address callInsn,
        ParseAPI::Function *&func,
        Address libLoad)
{

    // Check if we have a cached result
    std::tr1::unordered_map<Address,ParseAPI::Function*>::iterator cacheIter;
    cacheIter = cachedCallees.find(callInsn);
    if (cacheIter != cachedCallees.end()) {
        func = cacheIter->second; 
        if (DEBUG) fprintf(stderr, "\t\t findCallee: found cached callee function = %s\n", func->name().c_str());
        return true;
    }

    ParseAPI::Interproc pred;
    const ParseAPI::Block::edgelist & targetEdges = srcblock->targets();

    ParseAPI::Edge * edge;
    ParseAPI::Block::edgelist::const_iterator eIter;
    for (eIter = targetEdges.begin(); eIter != targetEdges.end(); ++eIter) {
        edge = *eIter;

        if (edge->type() != ParseAPI::CALL) {
            if(DEBUG) fprintf(stderr, "\t\t\t findCallee: discarded non-call edge to 0x%lx\n", edge->trg()->start());
            continue;
        }

        // Special case; found sink edge
        if (edge->sinkEdge()) {
            if (DEBUG) fprintf(stderr, "\t\t\t findCallee: call edge has unresovled control flow target; conservatively assume edge exists to callee\n");
            return true;
        }

        ParseAPI::Block * target = edge->trg();
        Address targetAddr = target->start();

       return findCalleeFunc(walker, callInsn, srcCS, func, libLoad, targetAddr); 
    }

    if (DEBUG) fprintf(stderr, "\t\t\t findCallee: reached end of function, returning false\n");
    return false;
}


