/* 
 * ROPStop is Copyright (c) 2012-2014 Barton P. Miller. Written by Emily R. Jacobson.
 *
 * This file is part of ROPStop.
 *
 * ROPStop is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * ROPStop is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with ROPStop.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <iostream>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

#include "PCProcess.h"

#include "initProcess.h"
#include "globals.h"

using namespace std;
using namespace Dyninst;

Architecture arch;
std::tr1::unordered_map<string, ParseAPI::SymtabCodeSource*> srcs;
std::tr1::unordered_map<string, ParseAPI::CodeObject*> objs;

ProcControlAPI::Process::ptr initProcess(int pid, char * bin_file, char * args, 
        char * stdin_arg, char * stdout_arg, char * stderr_arg, 
        char * func)
{
    ProcControlAPI::Process::ptr proc;

    if (pid != -1) {
        /* Attach to running process */
        if (VERBOSE || DEBUG) fprintf(stderr, "Attaching to process %d\n", pid);
        proc = ProcControlAPI::Process::attachProcess(pid);
        if (!proc) {
            if (DEBUG) fprintf(stderr, "attachProcess(%d) failed: %s\n", pid, ProcControlAPI::getLastErrorMsg());
            exit(1);
        } 
    } else {
        /* Create a new process */
        std::vector<std::string> local_argv;
        local_argv.push_back(bin_file);

        /* Gather arguments */ 
        if (!LITERAL) {
            if (args) {
                char * tmp;
                const char * delim = " ";
                tmp = strtok(args, delim);
                while (tmp != NULL) {
                    string tmpStr(tmp);
                    local_argv.push_back(tmpStr);
                    tmp = strtok(NULL, delim);
                }
            }
        } else {
            if (DEBUG) fprintf(stderr, "Using literal arguments\n");
            local_argv.push_back(args);  
        }

        /* Put together any necessary env stuff */
        std::vector<std::string> envp;

        /* Put together any necessary fd redirection */
        std::map<int,int> fds;
        /* stdin */ 
        if (stdin_arg) {
            int stdin_fd = open(stdin_arg, O_RDONLY);
            if (stdin_fd == -1) {
                perror("open");
                cerr << "stdin_arg is " << stdin_arg << endl;
                exit(1);
            }

            fds.insert(make_pair(stdin_fd, 0));
            if (DEBUG) fprintf(stderr, "Redirectings stdin from %s (fd %d)\n", stdin_arg, stdin_fd);
        }
        
        /* stdout */ 
        if (stdout_arg) {
            int stdout_fd = open(stdout_arg, O_WRONLY);
            if (stdout_fd == -1) {
                perror("open");
                cerr << "stdout_arg is " << stdout_arg << endl;
                exit(1);
            }

            fds.insert(make_pair(stdout_fd, 1));
            if (DEBUG) fprintf(stderr, "Redirectings stdout from %s (fd %d)\n", stdout_arg, stdout_fd);
        }
        
        /* stderr */ 
        if (stderr_arg) {
            int stderr_fd = open(stderr_arg, O_WRONLY);
            if (stderr_fd == -1) {
                perror("open");
                cerr << "stderr_arg is " << stderr_fd << endl;
                exit(1);
            }

            fds.insert(make_pair(stderr_fd, 2));
            if (DEBUG) fprintf(stderr, "Redirectings stderr from %s (fd %d)\n", stderr_arg, stderr_fd);
        }
        
        if (VERBOSE || DEBUG) fprintf(stderr, "Creating process from binary %s\n", bin_file);
        proc = ProcControlAPI::Process::createProcess(bin_file, local_argv, envp, fds);
        if (!proc) {
            if (DEBUG) fprintf(stderr, "createProcess(%s) failed: %s\n", bin_file, ProcControlAPI::getLastErrorMsg());
            exit(1);
        } 

        /* Update pid information */
        pid = proc->getPid();
        if (DEBUG) fprintf(stderr, "New PID is %d\n", pid);
    }

    // Update architecture info
    arch = proc->getArchitecture();

    /* Pause execution for a while */
    if (!proc->stopProc()) {
        if (DEBUG) fprintf(stderr, "proc->stopProc() failed\n");
        exit(1);
    }

    return proc;
}
