/* 
 * ROPStop is Copyright (c) 2012-2014 Barton P. Miller. Written by Emily R. Jacobson.
 *
 * This file is part of ROPStop.
 *
 * ROPStop is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * ROPStop is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with ROPStop.  If not, see <http://www.gnu.org/licenses/>.
 */

/* Includes */
#include <stdio.h>
#include <getopt.h>
#include <iostream>
#include <sstream>
#include <assert.h>
#include <stack>

#include "dyntypes.h"

/* ProcControl includes */
#include "PCProcess.h"
#include "PCErrors.h"

/* Stackwalker includes */
#include "walker.h"
#include "frame.h"
#include "basetypes.h"
#include "framestepper.h"
#include "steppergroup.h"
#include "swk_errors.h"

/* InstructionAPI includes */
#include "Instruction.h"
#include "InstructionDecoder.h"

/* ParseAPI includes */
#include "CFG.h"
#include "CodeObject.h"

/* SymtabAPI includes */
#include "Symtab.h"

using namespace std;
using namespace Dyninst;

#include "manualWalk.h"
#include "validRA.h"
#include "printers.h"
#include "globals.h"
#include "accessors.h"
#include "findCallee.h"
#include "validCall.h"
#include "globals.h"

bool manualWalkStack(Stackwalker::Walker * walker, 
        ProcControlAPI::Process::const_ptr proc,
        ProcControlAPI::Thread::ptr thread,
        MachRegisterVal val,
        std::stack<callLocation> & callstack, 
        InstructionAPI::Instruction::Ptr & syscall,
        Address & syscallAddr,
        std::string & error_message)
{
    
    if (VERBOSE || DEBUG) cerr << "\t manualWalkStack: completing a manual stackwalk @ PC: " << std::hex << val << "----------" << endl;
   
    std::vector< std::pair<Stackwalker::Frame,callLocation> > stackwalk;
    Stackwalker::WandererHelper wanderer(walker->getProcessState());
    Stackwalker::Frame initFrame(walker);
    if (!(walker->getInitialFrame(initFrame, thread->getLWP()))) {
        if (DEBUG) fprintf(stderr, "manualWalkStack: getInitialFrame failed: %s\n", Stackwalker::getLastErrorMsg());
        return false;
    } else {

        /* add the top frame to our stackwalk */
//        std::vector<Stackwalker::Frame> stackwalk;
        stackwalk.clear();
        InstructionAPI::Instruction::Ptr instr;
        callLocation curLocation((Address)initFrame.getRA(), getFrameFunc(walker, initFrame.getRA()), instr);
        stackwalk.push_back(std::make_pair(initFrame,curLocation));

        
        std::string lib;
        Dyninst::Offset offset;
        void * symtab;
        
        Stackwalker::Frame calleeFrame;
        Stackwalker::Frame callerFrame(walker);
        string calleeName, callerName;
            
        InstructionAPI::Instruction::Ptr insn;
        Address insn_addr;
        bool top = true;
        
        // Set calleeFrame initially
        calleeFrame = stackwalk.back().first;

        for (;;) {
        
            calleeFrame.getName(calleeName);
            callerFrame.getName(callerName);
           
            if (calleeName == "_dl_sysinfo_int80") {
                if (DEBUG) cerr << "\t\t manualWalkStack: missed bottom of stack detection (_dl_sysinfo_int80), returning true anyway" << endl;
                break;
            }

            if (calleeName == "start_thread") {
                if (DEBUG) cerr << "\t\t manualWalkStack: found start_thread; reached stack bottom" << endl;
                break;
            }

            if (calleeName == "__clone") {
                if (DEBUG) cerr << "\t\t manualWalkStack: found __clone; not walking further" << endl;
                break;
            }

            /* Check if we're still in loader code; if so, return true */
            if ((calleeFrame.getLibOffset(lib, offset, symtab))) {
                if (lib.find("ld-") != string::npos) {
                    if (DEBUG) fprintf(stderr, "\t\t manualWalkStack: still in loader code, skipping this stackwalk\n");
                    return true;
                }
            }

            if (!walker->walkSingleFrame(calleeFrame, callerFrame)) {
                if (Stackwalker::getLastError() == Stackwalker::err_stackbottom) {
                    if (DEBUG) fprintf(stderr, "manualWalkStack: %s\n", Stackwalker::getLastErrorMsg());
                    break;
                } else {
                    if (DEBUG) cerr << "\t manualWalkStack: walkSingleFrame failed (" << std::hex << calleeFrame.getRA() << "): " << Stackwalker::getLastErrorMsg() << endl;
                    if (DEBUG) cerr << "\t\t manualWalkStack: Callee: " << printFrameInfo(calleeFrame, true) << endl; 
                    if (DEBUG) cerr << "\t\t manualWalkStack: Caller: " << printFrameInfo(callerFrame, true) << endl;
                    return false;
                }
            }  

            

            if (DEBUG) fprintf(stderr, "\t\t manualWalkStack: calculated calleeFrame %s, checking for valid RA\n", printFrameInfo(calleeFrame, false).c_str());
            if (DEBUG) fprintf(stderr, "\t\t manualWalkStack: expected callerFrame %s, checking for valid RA\n", printFrameInfo(callerFrame, false).c_str());

            /* Check the state of this frame */
            if (!validRA(walker, proc, callerFrame.getRA(), insn, insn_addr)) {
                
                Stackwalker::location_t raloc = callerFrame.getRALocation();
                if (raloc.location == Stackwalker::loc_unknown) {
                    if (DEBUG) fprintf(stderr, "\t\t RA location is loc_unknown, skip this frame: %s\n", printFrameInfo(callerFrame, false).c_str());

                    // Still need to set the next callee frame
                    calleeFrame = callerFrame;

                    continue;
                }

                ParseAPI::Block * rablock = getFrameFunc(walker, callerFrame.getRA())->entry();
                if (rablock->start() == rablock->lastInsnAddr()) {
                    if (DEBUG) fprintf(stderr, "\t\t RA location in single-instruction function, skipping this frame: %s\n", printFrameInfo(callerFrame, false).c_str());
                    calleeFrame = callerFrame;

                    continue;
                }
           
                if (DEBUG) cerr << "\t\t manualWalkStack: frame has invalid height: " << printFrameInfo(calleeFrame, false) << ", caller RA " << callerFrame.getRA() << " is invalid" << endl;
                error_message = "Invalid frame: ";
                error_message.append(calleeName);
                return false;     
            } else {
                if (DEBUG || VERBOSE) cerr << "\t\t manualWalkStack: frame has valid height: " << printFrameInfo(calleeFrame, false) << endl;
            }

            if (top) {
                syscall = insn;
                syscallAddr = insn_addr;
                top = false;
            }
            
            callLocation loc(insn_addr, getFrameFunc(walker, insn_addr), insn);
            stackwalk.push_back(std::make_pair(callerFrame, loc));
            
            // Grab the next callee to walk out of 
            calleeFrame = stackwalk.back().first;
        }
        // Second pass: check for a valid call path
        if (VERBOSE || DEBUG) cerr << "\t manualWalkStack: verifying valid callstack----------" << endl;
        int numFrames = stackwalk.size();
        for (int index = 0; index < numFrames-1; index++) {
            calleeFrame = stackwalk[index].first;
            std::pair<Stackwalker::Frame,callLocation> callerP = stackwalk[index+1];
            callerFrame = callerP.first;
            insn = callerP.second.instr();
            insn_addr = callerP.second.addr();
            if (DEBUG) cerr << "\t manualWalkStack: checking for valid call " << printFrameInfo(callerFrame, false) << " -> " << printFrameInfo(calleeFrame, false) << " at 0x" << insn_addr << endl;
            if (!validCall(walker, insn, insn_addr, callerFrame, calleeFrame)) {
                error_message = "No valid call: ";
                error_message.append(callerName);
                error_message.append("->");
                error_message.append(calleeName);

                if (DEBUG || VERBOSE) cerr << "\t\t manualWalkStack: invalid caller->callee to " << printFrameInfo(calleeFrame, false) << endl;

                return false;            
            } else {
                if (VERBOSE || DEBUG) cerr << "\t\t manualWalkStack: found valid call to " << printFrameInfo(calleeFrame, false) << endl;
            }
        } 
    }
    
    int numFrames = stackwalk.size();
    for (int index = numFrames-1; index >= 0; index--) {
        callstack.push(stackwalk[index].second);
    }

    return true;
}

