/* 
 * ROPStop is Copyright (c) 2012-2014 Barton P. Miller. Written by Emily R. Jacobson.
 *
 * This file is part of ROPStop.
 *
 * ROPStop is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * ROPStop is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with ROPStop.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "PCProcess.h"
#include "dyntypes.h"
#include "ProcessSet.h"
#include "Instruction.h"

#include "monitorProcess.h"
#include "validPC.h"
#include "globals.h"
#include "initProcess.h"
#include "setupProcess.h"

/* manualWalk includes */
#include "manualWalk.h"
#include "accessors.h"
#include "printers.h"
//#include "printStack.h"

using namespace std;
using namespace Dyninst;

bool ATTACKED = false;
bool FIRST_TIME_THROUGH = true;

/* Monitor a process via various callbacks */
bool monitorProcess()
{
    ProcControlAPI::Process::ptr proc;
            
    ProcControlAPI::ThreadPool::iterator tIter;
    ProcControlAPI::Thread::ptr thread;
    bool procTerminate;
    bool threadTerminate;
                    
    MachRegisterVal pc;

    ParseAPI::SymtabCodeSource * cs;
    ParseAPI::CodeObject * co;

    std::string libName;
    Address libLoad;
    unsigned libSize;
                    
    Address relAddr;
    std::set<ParseAPI::CodeRegion *> regions;
    std::string error_message;

    set<PW_pair> terminated;
    set<PW_pair>::iterator mIter, mIter2;
    set<PW_pair>::iterator termIter;
    
    /* Wait for process to terminate */
    while (monitored.size()) {
        bool block = true;
        ProcControlAPI::Process::handleEvents(block);

        /* Handle new stuff in our AS... */
        if (NEW_EXEC) {
            if (DEBUG) fprintf(stderr, "\t monitorProcess: process %d exec'd\n", NEW_EXEC->getPid());
            
            /* Setup walker */
            Stackwalker::Walker* walker;
            if (!setupProcess(NEW_EXEC, &walker, ANALYSIS)) {
                exit(1);
            }

            /* Remove information about previous version (pre-exec) */
            PW_pair toDelete;
            bool found = false; 
            for (mIter = monitored.begin(); mIter != monitored.end(); ++mIter) {
                PW_pair curPair = *mIter;
                proc = curPair.first;
                if (proc == NEW_EXEC) {
                    toDelete = curPair;
                    found = true;
                    break;
                }
            }
            
            if (found) monitored.erase(toDelete);
            monitored.insert(make_pair(NEW_EXEC, walker));

            NEW_EXEC = ProcControlAPI::Process::ptr();
        }

        if (NEW_PROCESS) {
            if (DEBUG) fprintf(stderr, "\t monitorProcess: there's a new process (%d) to monitor!\n", NEW_PROCESS->getPid());
            
            proc = NEW_PROCESS;

            /* Setup walker */
            Stackwalker::Walker * walker;
            if (!setupProcess(proc, &walker, ANALYSIS)) {
                if (DEBUG) fprintf(stderr, "\t monitorProcess: process %d setup failed, exiting\n", proc->getPid());
                exit(1);
            } 
            monitored.insert(make_pair(proc, walker));

            /* Reset new process for next fork */
            NEW_PROCESS = ProcControlAPI::Process::ptr();
        }

        if (NEW_THREAD) {
            if (DEBUG) fprintf(stderr, "\t monitorProcess: there's a new thread (%d) to monitor!\n", NEW_THREAD->getLWP());

            NEW_THREAD->setSyscallMode(true);
            NEW_THREAD = ProcControlAPI::Thread::ptr();
        }

        /* Main processing loop */
        procTerminate = false;
        for (mIter = monitored.begin(); mIter != monitored.end(); ++mIter) {
            PW_pair curPair = *mIter;
            proc = curPair.first;
            Stackwalker::Walker * walker = curPair.second;

            if (proc->isTerminated()) {
                if (DEBUG) fprintf(stderr, "monitorProcess: process %d has terminated\n", proc->getPid());
                terminated.insert(curPair);
                continue;
            }
            if (!proc->hasStoppedThread()) {
                if (!proc->continueProc()) {
                    if (DEBUG) fprintf(stderr, "proc->continueProc failed: %s\n", ProcControlAPI::getLastErrorMsg());
                    break;
                }
                continue;
            }

            if (DEBUG) fprintf(stderr, "monitorProcess: process %d has at least one thread stopped, checking threads\n", proc->getPid());

            ProcControlAPI::ThreadPool & threads = proc->threads();
            for (tIter = threads.begin(); tIter != threads.end(); ++tIter) {     
                thread = *tIter;
                threadTerminate = false;

                if (!thread->isStopped()) {
                    if (DEBUG) fprintf(stderr, "\t monitorProcess: thread %d is running\n", thread->getLWP());
                    continue;
                } else {
                    if (DEBUG) fprintf(stderr, "\t monitorProcess: thread %d is stopped\n", thread->getLWP());

#if CHECKER
                    std::stack<callLocation> callstack; 

                    /* Get current PC */
                    if (!thread->getRegister(MachRegister::getPC(proc->getArchitecture()), pc)) {
                        if (DEBUG) fprintf(stderr, "\t monitorProcess: thread->getRegister(pc) failed\n");
                        procTerminate = true;
                        break;
                    }

                    
                    /* What system call is this? */
                    MachRegisterVal eax;
                    thread->getRegister(EAX, eax);

                    /* Gather ParseAPI data */
                    cs = NULL;
                    co = NULL;

                    if (!getLibraryAndSize(walker, pc, libName, libLoad, libSize)) {
                        if (DEBUG) fprintf(stderr, "monitorProcess: getLibraryAndSize failed\n");
                        return false;
                    } else {

                        if (DEBUG) fprintf(stderr, "monitorProcess: library \"%s\" loaded at 0x%lx\n", libName.c_str(), libLoad);

                        // Skip the loader
                        size_t str = libName.find("ld-");
                        if (str != string::npos) {
                            if (DEBUG) fprintf(stderr, "\t monitorProcess: skipping loader code\n");
                            
                            if (DEBUG) fprintf(stderr, "Thread %d continuing\n", thread->getLWP());

                            /* Let the process continue for a while */ 
                            if (!thread->continueThread()) {
                                if (DEBUG) fprintf(stderr, "thread->continueThread failed: %s\n", ProcControlAPI::getLastErrorMsg());
                                break;
                            }
                            continue;
                        }
                        cs = getCodeSource(walker, libName, libLoad, libSize);
                        if (!cs) {
                            if (DEBUG) fprintf(stderr, "monitorProcess: getCodeSource failed\n");
                            return false;
                        }
                        co = getCodeObject(walker, libName, libLoad, libSize);                        
                        if (!co) {
                            if (DEBUG) fprintf(stderr, "monitorProcess: getCodeObject failed\n");
                        }
                    }
                    
                    // Find relevant code region(s)
                    relAddr = getRelAddr(pc, libName, libLoad);
                    regions = getCodeRegions(cs, relAddr);
                   
                    if (VERBOSE) fprintf(stderr, "monitorProcess: Checking for program conformance at 0x%lx (thread %d)=================================\n", pc, thread->getLWP());
                    if (VERBOSE) fprintf(stderr, "\t monitorProcess: system call %d\n", (int)eax);

                    /* Verify PC validity;
                     * - Verify that PC points to an executable portion of memory
                     * - ORIG_INSTR global controls whether we also check that PC points to valid instruction
                     */

                    InstructionAPI::Instruction::Ptr syscall;
                    Address syscallAddr;
                    InstructionAPI::Instruction::Ptr instr;
                    ParseAPI::Block * block;
                    if (!validPC(walker, thread->getLWP(), pc, co, regions, libName, libLoad, libSize, instr, block)) {
                        if (VERBOSE) fprintf(stderr, "\t monitorProcess: found invalid PC, terminating process!\n");
                        threadTerminate = true;
                        ATTACKED = true;
                    } else {
                        if (VERBOSE) fprintf(stderr, "\t monitorProcess: found valid PC, checking callstack\n");

                        /* Verify callstack validity */
                        if (!manualWalkStack(walker, proc, thread, pc, callstack, syscall, syscallAddr, error_message)) {
                            if (VERBOSE) fprintf(stderr, "\t monitorProcess: found invalid callstack (%s), terminating process!\n", error_message.c_str());

                            threadTerminate = true;
                            ATTACKED = true;
                        } else {
                            if (VERBOSE) fprintf(stderr, "\t monitorProcess: found valid callstack, continuing thread %d\n", thread->getLWP());
                        }
                    }
#endif

#if !CONTINUE
                    /* Terminate the process if we've detected an attack */
                    if (threadTerminate) {
                        procTerminate = true;
                        break;
                    } else {
                        thread->continueThread();
                    }
#else
                    thread->continueThread();
#endif
                }
            }
            /* If we've found a bad state, terminate everything */
            if (procTerminate) {
        
                /* Terminate all other processes, also */
                fprintf(stderr, "Terminating all processes\n");
#if !CONTINUE
                for (mIter2 = monitored.begin(); mIter2 != monitored.end(); ++mIter2) {
                    ProcControlAPI::Process::ptr curProc = (*mIter2).first;
                    
                    fprintf(stderr, "Terminating %d\n", curProc->getPid());
                    if (!curProc->terminate()) {
                       fprintf(stderr, "Terminating %d failed: %s\n", 
                               curProc->getPid(),
                               ProcControlAPI::getLastErrorMsg());
                   } else {
                    fprintf(stderr, "Terminated %d\n", 
                            curProc->getPid());
                   } 
                }
                fprintf(stderr, "Finished terminating all processes.\n");
                return !ATTACKED;
#else
                fprintf(stderr, "\t Temporary debug step: will allow process to continue\n");
                procTerminate = false;

#endif
            }

            if (DEBUG) fprintf(stderr, "Process %d continuing\n", proc->getPid());

            /* Let the process continue for a while */ 
            if (!proc->continueProc()) {
                if (DEBUG) fprintf(stderr, "proc->continueProc failed: %s\n", ProcControlAPI::getLastErrorMsg());
                break;
            }
        } // end for(mIter) loop
        
        /* Remove terminated processes */
        for (termIter = terminated.begin(); termIter != terminated.end(); ++termIter) {
            monitored.erase(*termIter);    
        }
        terminated.clear();
    }
    if (DEBUG) fprintf(stderr, "No processes remain to be monitored; return\n");

    return !ATTACKED;
}
