/* 
 * ROPStop is Copyright (c) 2012-2014 Barton P. Miller. Written by Emily R. Jacobson.
 *
 * This file is part of ROPStop.
 *
 * ROPStop is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * ROPStop is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with ROPStop.  If not, see <http://www.gnu.org/licenses/>.
 */

/* Includes */
#include <stdio.h>
#include <getopt.h>
#include <iostream>
#include <sstream>
#include <fstream>
#include <set>

#include <sys/time.h>
#include <sys/resource.h>

#include <sys/ptrace.h>
#include <sys/types.h>
#include <sys/wait.h>
//#include <linux/user.h>
#include <sys/user.h>

#include "dyntypes.h"

/* ProcControl includes */
#include "PCProcess.h"
#include "PCErrors.h"
#include "Event.h"

/* Stackwalker includes */
#include "walker.h"
#include "frame.h"
#include "basetypes.h"
#include "framestepper.h"
#include "steppergroup.h"
#include "swk_errors.h"

/* SymtabAPI includes */
#include "Module.h"

/* Local includes */
#include "initProcess.h"
#include "setupProcess.h"
#include "monitorProcess.h"
#include "globals.h"
#include "loadKnownLibraries.h"
#include "Timer.h"
#include "printers.h"

using namespace std;
using namespace Dyninst;

void usage(char* arg0)
{
    cerr << "Usage: " << arg0 << "\n"
        << "\t -b <binary> [-a <args>]" << endl
        << "\t OR -p <pid>]" << endl
        << "\t [-f <func>]" << endl
        << "\t [-s (use AnalysisStepper only)]" << endl
        << "\t [-c (config file)" << endl
        << "\t [-r (verify PC points to orig instr)]" << endl
        << "\t [-l (take args as one literal)]" << endl
        << "\t [-v (verbose)]" << endl;
}

/* Input option globals */
int CUR_PID = -1;
std::string BIN_FILE;
std::string ARGS;
bool ANALYSIS = false;
char * FUNC = NULL;
std::string STDIN_ARG, STDOUT_ARG, STDERR_ARG;
bool ALL_SYSCALLS = true;
bool VERBOSE = false;
bool DEBUG = false;
bool LITERAL = false;
bool ORIG_INSTR = true;
std::string CONFIG_FILE;
set<PW_pair> monitored;

/*
 * Parse input options
 */
void parse_options(int argc, char** argv)
{
    static struct option long_options[] = {
        {"help", no_argument, 0, 'h'}
    };

    int ch;
    int option_index = 0;

    while ((ch = getopt_long(argc, 
                    argv,
                    "b:p:a:f:i:e:o:slrc:vdh -",
                    long_options,
                    &option_index)) != -1)
    {
        switch(ch) {
            case 'p':
                CUR_PID = atoi(optarg);
                cerr << "Setting CUR_PID to " << CUR_PID << "; optarg = " << optarg << endl;
                break;
            case 'b':
                BIN_FILE = optarg;
                break;
            case 'a':
                ARGS = std::string(optarg);
                break;
            case 'f':
                FUNC = optarg;
                break;
            case 'i':
                STDIN_ARG = std::string(optarg);
                break;
            case 'e':
                STDERR_ARG = std::string(optarg);
                break;
            case 'o':
                STDOUT_ARG = std::string(optarg);
                break;
            case ' ':
            case '-':
                break;
            default:
                printf("Illegal option %c\n", ch);
            case 'h':
                usage(argv[0]);
                exit(0);
            case 's':
                ANALYSIS = true;
                break;
            case 'c':
                CONFIG_FILE = optarg;
                cerr << "Setting config_file to " << CONFIG_FILE << endl;
                break;
            case 'v':
                VERBOSE = true;
                break;
            case 'd': 
                DEBUG = true;
                break;
            case 'l':
                LITERAL = true;
                break;
            case 'r':
                ORIG_INSTR = true;
                break;
        }
    }

    if (CUR_PID == -1 && BIN_FILE == "" && CONFIG_FILE == "") {
        usage(argv[0]);
        exit(1);
    }
}

int main(int argc, char** argv)
{
    timer * t = new timer();
    t->start();

    parse_options(argc, argv);

    fprintf(stderr, "ropstop called with\n"
            "\t BINARY = \"%s\"\n"
            "\t [ARGS = \"%s\"]\n"
            "\t [STDIN = \"%s\"]\n"
            "\t [STDOUT = \"%s\"]\n"
            "\t [STDERR = \"%s\"]\n"
            "\t [FUNC = \"%s\"]\n"
            "\t [PID = \"%d\"]\n", 
            BIN_FILE.c_str(), ARGS.c_str(), 
            STDIN_ARG.c_str(), STDOUT_ARG.c_str(), 
            STDERR_ARG.c_str(), FUNC, 
            CUR_PID);
    if (VERBOSE) fprintf(stderr, "\t VERBOSE mode\n");
    if (LITERAL) fprintf(stderr, "\t LITERAL args mode\n");
    if (ORIG_INSTR) fprintf(stderr, "\t PC: verify ORIG_INSTR mode\n");
    if (ALL_SYSCALLS) fprintf(stderr, "\t VERIFY ALL_SYSCALLS mode\n");

    if (CONFIG_FILE != "") {
        /* If we have a config file, we're going to batch these */
        ifstream file;
        cerr << "Opening " << CONFIG_FILE << endl;
        file.open(CONFIG_FILE.c_str(), ifstream::in);
        if (!file.is_open()) {
            cerr << "File open failed" << endl;
            exit(1);
        }
        while(1) {
            string line;
            getline(file, line);
            if (file.eof()) break;
            size_t index = 0;

            cerr << "Input line: " << file << endl;

            index = line.find(' ');
            BIN_FILE = line.substr(0, index); 
            line = line.substr(index + 1);
            cerr << "\t bin file: " << BIN_FILE << endl;
            unsigned int tmp = 0;
            while(1) {
                tmp++;
                if (tmp > 10) exit(1);
                index = line.find(' ');
                if (index == std::string::npos) break;

                std::string tmp = line.substr(0, index);
                line = line.substr(index + 1);
                cerr << "\t tmp line: " << tmp << endl;
                std::string next;

                if (tmp == "-i" ||
                        tmp == "-e" ||
                        tmp == "-o") {
                    index = line.find(' ');
                    next = line.substr(0, index);
                    line = line.substr(index + 1);
                }

                if (tmp == "-i") {
                    STDIN_ARG = next;
                    cerr << "\t stdin: " << STDIN_ARG << endl;
                }
                else if (tmp == "-o") {
                    STDOUT_ARG = next;
                    cerr << "\t stdout: " << STDOUT_ARG << endl;
                }
                else if (tmp == "-e") {
                    STDERR_ARG = next;
                    cerr << "\t stderr: " << STDERR_ARG << endl;
                }
                else if (tmp == "--") {
                    ARGS = line;
                    cerr << "\t args: " << ARGS << endl;
                    break;
                }

            }
            /* Load known libraries ahead of time */
            if (BIN_FILE.size()) {
                if (!loadKnownLibraries((char *) BIN_FILE.c_str())) {
                    if (DEBUG) fprintf(stderr, "loadKnown libraries failed\n");
                }
            }

            cerr << "\t init proc!" << endl;
            
            printRusage();
            
            /* Setup process */
            ProcControlAPI::Process::ptr proc = initProcess(CUR_PID,
                    BIN_FILE.size() ? (char *) BIN_FILE.c_str() : NULL,
                    ARGS.size() ? (char *) ARGS.c_str() : NULL,
                    STDIN_ARG.size() ? (char *)STDIN_ARG.c_str() : NULL,
                    STDOUT_ARG.size() ? (char *)STDOUT_ARG.c_str() : NULL,
                    STDERR_ARG.size() ? (char *)STDERR_ARG.c_str() : NULL,
                    FUNC);

            /* Setup walker */
            Stackwalker::Walker * walker;
            if (!setupProcess(proc, &walker, ANALYSIS)) {
                fprintf(stderr, "Process %d setup failed, exiting\n", proc->getPid());
                exit(1);
            }

            printRusage();

            monitored.insert(make_pair(proc, walker));

            /* Monitor process */
            bool status = monitorProcess();

            fprintf(stderr, "All processes have terminated\n");
            if (status) fprintf(stderr, "\t No exploits were detected.\n");
            else fprintf(stderr, "\t An exploit was detected and the program was prematurely terminated.\n");

        }

        t->stop();
        fprintf(stderr, "Total: u: %f, s: %f, w: %f (%d system calls)\n", t->usecs(), t->ssecs(), t->wsecs(), COUNTER);

    } else {
        /* Load known libraries ahead of time */
        if (BIN_FILE.size()) {
            if (!loadKnownLibraries((char *) BIN_FILE.c_str())) {
                if (DEBUG) fprintf(stderr, "loadKnown libraries failed\n");
            }
        }

        t->stop();
        fprintf(stderr, "Elapsed: u: %f, s: %f, w: %f\n", t->usecs(), t->ssecs(), t->wsecs());
        t->start();

        /* Setup process */
        ProcControlAPI::Process::ptr proc = initProcess(CUR_PID, 
                BIN_FILE.size() ? (char *) BIN_FILE.c_str() : NULL,
                ARGS.size() ? (char *) ARGS.c_str() : NULL,
                STDIN_ARG.size() ? (char *)STDIN_ARG.c_str() : NULL,
                STDOUT_ARG.size() ? (char *)STDOUT_ARG.c_str() : NULL,
                STDERR_ARG.size() ? (char *)STDERR_ARG.c_str() : NULL,
                FUNC);

        printRusage();

        /* Setup walker */
        Stackwalker::Walker * walker;
        if (!setupProcess(proc, &walker, ANALYSIS)) {
            fprintf(stderr, "Process %d setup failed, exiting\n", proc->getPid());
            exit(1);
        }

        printRusage();

        monitored.insert(make_pair(proc, walker));

        /* Monitor process */
        bool status = monitorProcess();

        fprintf(stderr, "All processes have terminated\n");
        if (status) fprintf(stderr, "\t No exploits were detected.\n");
        else fprintf(stderr, "\t An exploit was detected and the program was prematurely terminated.\n");

        t->stop();
        fprintf(stderr, "Total: u: %f, s: %f, w: %f (%d system calls)\n", t->usecs(), t->ssecs(), t->wsecs(), COUNTER);
    }

    printRusage();

    return 0;
}
