/* 
 * ROPStop is Copyright (c) 2012-2014 Barton P. Miller. Written by Emily R. Jacobson.
 *
 * This file is part of ROPStop.
 *
 * ROPStop is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * ROPStop is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with ROPStop.  If not, see <http://www.gnu.org/licenses/>.
 */

/* Includes */
#include <stdio.h>
#include <getopt.h>
#include <iostream>
#include <sstream>
#include <assert.h>
#include "dyntypes.h"
#include <tr1/unordered_set>

/* ProcControl includes */
#include "PCProcess.h"
#include "PCErrors.h"

/* Stackwalker includes */
#include "walker.h"
#include "frame.h"
#include "basetypes.h"
#include "framestepper.h"
#include "steppergroup.h"
#include "swk_errors.h"

/* InstructionAPI includes */
#include "Instruction.h"
#include "InstructionDecoder.h"

/* ParseAPI includes */
#include "CFG.h"
#include "CodeObject.h"

/* SymtabAPI includes */
#include "Symtab.h"

using namespace std;
using namespace Dyninst;

#include "manualWalk.h"
#include "printers.h"
#include "globals.h"
#include "accessors.h"
#include "findCallee.h"
#include "validCall.h"

std::tr1::unordered_map<Address,std::set<ParseAPI::Function*> > cachedCallEdges;
std::tr1::unordered_map<ParseAPI::Function*,std::set<ParseAPI::Function*> > cachedTailCalls; 
std::tr1::unordered_map<Address, std::pair<InstructionAPI::Instruction::Ptr, ParseAPI::Function*> > cachedManualInstructions; // map from address to (insn@addr, lastvalidinsn)
std::tr1::unordered_map<Address, std::set<ParseAPI::Function*> > cachedFrameFuncs;

bool getFuncs(Stackwalker::Walker * walker,
        Address calleeFrameInsn,
        std::set<ParseAPI::Function*> & funcs)
{
    // TODO: Cache here!
    std::tr1::unordered_map<Address, std::set<ParseAPI::Function*> >::iterator iter;
    iter = cachedFrameFuncs.find(calleeFrameInsn);
    if (iter != cachedFrameFuncs.end()) {
        if (DEBUG) fprintf(stderr, "\t\t\t getFuncs: found cached set of funcs\n");
        funcs = iter->second;
        return true;
    }

    std::set<ParseAPI::Function*> retSet;

    // What function(s) is associated with the callee frame?
    ParseAPI::SymtabCodeSource * cs = NULL;
    ParseAPI::CodeObject * co = NULL;

    std::string libName;
    Address libLoad;
    unsigned libSize;
    if (!getLibraryAndSize(walker, calleeFrameInsn, libName, libLoad, libSize)) {
        if (DEBUG) fprintf(stderr, "\t\t\t getFuncs: getLibraryAndSize failed\n");
        return false;
    } else {
        if (DEBUG) fprintf(stderr, "\t\t\t getFuncs: library \"%s\" loaded at 0x%lx\n", libName.c_str(), libLoad);

        cs = getCodeSource(walker, libName, libLoad, libSize);
        if (!cs) {
            if (DEBUG) fprintf(stderr, "\t\t getFuncs: getCodeSource failed\n");
            return false;
        }

        co = getCodeObject(walker, libName, libLoad, libSize);                        
        if (!co) {
            if (DEBUG) fprintf(stderr, "\t\t\t getFuncs: getCodeObject failed\n");
            return false;
        }
    }

    // Get CodeRegion
    Address relAddr = getRelAddr(calleeFrameInsn, libName, libLoad);
    std::set<ParseAPI::CodeRegion *> regions = getCodeRegions(cs, relAddr);

    if (DEBUG) fprintf(stderr, "\t\t\t getFuncs: retrieving functions at relative address 0x%lx\n", relAddr);

    if (regions.size() == 0) {
        if (DEBUG) fprintf(stderr, "\t\t\t getFuncs: getCodeRegions failed\n");
        return false;
    }

    std::set<ParseAPI::CodeRegion *>::iterator rIter;
    std::set<ParseAPI::Function*> calleeFrameFuncs;
    std::set<ParseAPI::Function*>::iterator cIter;
    ParseAPI::CodeRegion * cr;
    for (rIter = regions.begin(); rIter != regions.end(); ++rIter) {
        cr = *rIter;
        co->findFuncs(cr, relAddr, funcs);
    }

    // Cache
    cachedFrameFuncs.insert(make_pair(calleeFrameInsn, funcs));

    return true;
}

bool isMatch(ParseAPI::Function * targetFunc, std::set<ParseAPI::Function *> funcSet)
{

    return (funcSet.find(targetFunc) != funcSet.end());
}

// Is there a (series of) tail calls from targetFunc to calleeFrameFuncs?
// targetFunc is the target of the original call instruction
bool validTailCall(Stackwalker::Walker * walker,
        ParseAPI::Function * targetFunc,
        std::set<ParseAPI::Function *> calleeFrameFuncs)
{
    // TODO: find possible tail-call-only path
    if (DEBUG) fprintf(stderr, "\t\t\t validTailCall: looking for a tail call from call target %s to one of the calleeFrameFuncs\n", 
            targetFunc->name().c_str());
    
    const ParseAPI::Function::blocklist & targBlocks = targetFunc->blocks();
    ParseAPI::Function::blocklist::const_iterator tIter;
    ParseAPI::Block::edgelist::const_iterator eIter;
    ParseAPI::Edge * edge;
    for (tIter = targBlocks.begin(); tIter != targBlocks.end(); ++tIter) {
        const ParseAPI::Block::edgelist & targetEdges = (*tIter)->targets();
        for (eIter = targetEdges.begin(); eIter != targetEdges.end(); ++eIter) {
            edge = *eIter;
            if (DEBUG) fprintf(stderr, "\t\t\t\t validTailCall: checking edge 0x%lx-0x%lx, type=%s, interproc=%d\n", 
                    edge->src()->lastInsnAddr(),
                    edge->trg()->start(), 
                    ParseAPI::format(edge->type()).c_str(),
                    edge->interproc()? 1 : 0);
            if (edge->interproc() && 
                    ( (edge->type() == ParseAPI::DIRECT ) || (edge->type() == ParseAPI::INDIRECT) )) {
                
                ParseAPI::SymtabCodeSource * cs = NULL;
                ParseAPI::CodeObject * co = NULL;

                std::string libName;
                Address libLoad;
                unsigned libSize;

                Address tailInsn = edge->src()->lastInsnAddr();

                if (!getLibraryAndSize(walker, tailInsn, libName, libLoad, libSize)) {
                    if (DEBUG) fprintf(stderr, "\t\t\t\t validTailCall: getLibraryAndSize failed\n");
                    return false;
                } else {
        
                    if (DEBUG) fprintf(stderr, "\t\t\t\t validTailCall: library \"%s\" loaded at 0x%lx\n", libName.c_str(), libLoad);
                    
                    cs = getCodeSource(walker, libName, libLoad, libSize);
                    if (!cs) {
                        if (DEBUG) fprintf(stderr, "\t\t validCall: getCodeSource failed\n");
                        return false;
                    }

                    co = getCodeObject(walker, libName, libLoad, libSize);                        
                    if (!co) {
                        if (DEBUG) fprintf(stderr, "\t\t validCall: getCodeObject failed\n");
                        return false;
                    }
                }

                ParseAPI::Function * tailTargetFunc = NULL;
                if (!findTailCallee(walker, edge->src(), edge, cs, tailTargetFunc, libLoad)) {
                    fprintf(stderr, "\t\t\t\t validTailCall: findTailCallee failed\n");
                    continue;
                } 

                if (!tailTargetFunc) {
                    // There was unresolved control flow
                    return true;
                }

                if (isMatch(tailTargetFunc, calleeFrameFuncs)) {
                    if (DEBUG) fprintf(stderr, "\t\t\t\t validTailCall: found tail call target in calleeFrameFuncs\n");
                    return true;
                }

            }
        }
    }
    return false;
}

bool manualFindCalleeAddress(InstructionAPI::Instruction::Ptr insn,
        Address callInsn,
        std::set<ParseAPI::Function *> calleeFrameFuncs,
        Address & callTargetAddr)
{
    if (DEBUG) fprintf(stderr, "\t\t\t manualFindCalleeAddress called: 0x%lx: %s\n",
            callInsn, insn->format().c_str());
    InstructionAPI::Expression::Ptr cft = insn->getControlFlowTarget();
    if (!cft) {
        if (DEBUG) fprintf(stderr, "\t\t\t manualFindCalleeAddress: getControlFlowTarget failed\n");
        callTargetAddr = -1;
        return false;
    }
    cft->bind(InstructionAPI::RegisterAST::Ptr(new InstructionAPI::RegisterAST(MachRegister::getPC(Dyninst::Arch_x86))).get(),
            InstructionAPI::Result(InstructionAPI::u32,callInsn));
    InstructionAPI::Result result = cft->eval();
    if (!result.defined) {
        if (DEBUG) fprintf(stderr, "\t\t\t manualFindCalleeAddress: unresolved control flow target\n");
        // Could not find a call target
        callTargetAddr = -1;
        return true;
    } 
    Address callTarget = result.convert<Address>();
    if (DEBUG) fprintf(stderr, "\t\t\t manualFindCalleeAddress: CFT = 0x%lx\n", callTarget);
    callTargetAddr = callTarget;
    return true;
}

void addToCache(Address callInsn, 
        std::set<ParseAPI::Function*> calleeFrameFuncs)
{
    cachedCallEdges.insert(make_pair(callInsn, calleeFrameFuncs));
}

bool validCall(Stackwalker::Walker * walker,
        InstructionAPI::Instruction::Ptr insn,
        Address callInsn,
        Stackwalker::Frame callerFrame,
        Stackwalker::Frame calleeFrame)
{
    string calleeName, callerName;

    /* Check that there exists a valid call between caller and callee at the previous insn */

    // What function(s) is associated with the callee Frame? 
    std::set<ParseAPI::Function *> calleeFrameFuncs;
    if (DEBUG) fprintf(stderr, "\t\t validCall: getting calleeFrameFuncs at 0x%lx\n", calleeFrame.getRA()-1);
    if (!getFuncs(walker, calleeFrame.getRA()-1, calleeFrameFuncs)) {
        if (DEBUG) fprintf(stderr, "\t\t validCall: getFuncs failed at 0x%lx\n", calleeFrame.getRA()-1);
        return false;
    }

    if (calleeFrameFuncs.size() == 0) {
        if (DEBUG) fprintf(stderr, "\t\t validCall: could not find function(s) at 0x%lx\n", calleeFrame.getRA()-1);
        std::string calleeLibName;
        Address calleeLibLoad;
        unsigned calleeLibSize;
        // We don't used calleeFrame.getRA()-1 here because we need to search only for *aligned* instructions.
        // Assume that RA will always have been parsed because we did a straight-line parse that included a few post-call instructions
        // to fill in the entire gap region.
        if (!getLibraryAndSize(walker, calleeFrame.getRA(), calleeLibName, calleeLibLoad, calleeLibSize)) {
            if (DEBUG) fprintf(stderr, "\t\t validCall: getLibraryAndSize (calleeFrame) failed\n");
            return false;
        }

        Address calleeRelAddr = getRelAddr(calleeFrame.getRA(), calleeLibName, calleeLibLoad);
        std::tr1::unordered_map<Address, std::pair<InstructionAPI::Instruction::Ptr, ParseAPI::Function*> >::iterator findManual;
        findManual = cachedManualInstructions.find(calleeRelAddr);
        if (findManual == cachedManualInstructions.end()) {
            if (DEBUG) fprintf(stderr, "\t\t validCall: could not find cached instr here either\n");
            // return false;

            // One last thing: what if we're in a single-instruction function?
            if (DEBUG) fprintf(stderr, "\t\t validCall: getting calleeFrameFuncs at 0x%lx\n", calleeFrame.getRA());
            if (!getFuncs(walker, calleeFrame.getRA(), calleeFrameFuncs)) {
                if (DEBUG) fprintf(stderr, "\t\t validCall: getFuncs failed at 0x%lx\n", calleeFrame.getRA());
                return false;
            } else {
                if (DEBUG) fprintf(stderr, "\t\t validCall: found calleeFrameFuncs at 0x%lx\n", calleeFrame.getRA());
            }
        } else {
            // Grab the cached function associated with this address
            ParseAPI::Function * func = findManual->second.second;
            calleeFrameFuncs.insert(func);
        }
    }

    // Cache!
    // Do we already know there's a valid call edge from callInsn to calleeFrameFuncs? 
    std::tr1::unordered_map<Address, std::set<ParseAPI::Function*> >::iterator iter;
    iter = cachedCallEdges.find(callInsn);
    if (iter != cachedCallEdges.end()) {
        if (iter->second == calleeFrameFuncs) {
            if (DEBUG) fprintf(stderr, "\t\t validCall: found cached call edge at 0x%lx\n", callInsn);
            return true;
        }
    } 
        
    ParseAPI::SymtabCodeSource * cs = NULL;
    ParseAPI::CodeObject * co = NULL;

    std::string libName;
    Address libLoad;
    unsigned libSize;
    if (!getLibraryAndSize(walker, callInsn, libName, libLoad, libSize)) {
        if (DEBUG) fprintf(stderr, "\t\t validCall: getLibraryAndSize failed\n");
        return false;
    } else {
                    
        if (DEBUG) fprintf(stderr, "\t\t validCall: library \"%s\" loaded at 0x%lx\n", libName.c_str(), libLoad);
        
        cs = getCodeSource(walker, libName, libLoad, libSize);
        if (!cs) {
            if (DEBUG) fprintf(stderr, "\t\t validCall: getCodeSource failed\n");
            return false;
        }

        co = getCodeObject(walker, libName, libLoad, libSize);                        
        if (!co) {
            if (DEBUG) fprintf(stderr, "\t\t validCall: getCodeObject failed\n");
            return false;
        }
    }

    // Find relevant code region(s)
    Address relAddr = getRelAddr(callInsn, libName, libLoad);
    std::set<ParseAPI::CodeRegion *> regions = getCodeRegions(cs, relAddr);
    
    if (DEBUG) fprintf(stderr, "\t\t validCall: investigating call 0x%lx (relative address = 0x%lx)\n", callInsn, relAddr);

    if (regions.size() == 0) {
        return false;
    }

    ParseAPI::CodeRegion * cr;
    std::set<ParseAPI::CodeRegion *>::iterator rIter;
    std::set<ParseAPI::Block *>::iterator bIter;
    ParseAPI::Block * srcblock;
    ParseAPI::Function * targetFunc;
    for (rIter = regions.begin(); rIter != regions.end(); ++rIter) {
        cr = *rIter; 

        std::set<ParseAPI::Block *> srcblocks;
        if (!co->findBlocks(cr, relAddr, srcblocks)) {
            if (DEBUG) fprintf(stderr, "\t\t validCall: could not find any caller blocks at 0x%lx (relative = 0x%lx) [%s]\n", callInsn, relAddr, cs->getSymtabObject()->name().c_str());
            continue;
        } else {
            for (bIter = srcblocks.begin(); bIter != srcblocks.end(); ++bIter) {
                srcblock = *bIter;

                std::vector<ParseAPI::Function*> srcFuncs;
                srcblock->getFuncs(srcFuncs);
                if (srcFuncs.size()) {
                    callerName = srcFuncs[0]->name();
                }
                               
                // What function is the target of this call edge?
                if (DEBUG) fprintf(stderr, "\t\t validCall: checking caller block 0x%lx-0x%lx\n", srcblock->start(), srcblock->lastInsnAddr());
                targetFunc = NULL;
                if (!findCallee(walker, srcblock, cs, relAddr, targetFunc, libLoad)) {
                    if (DEBUG) fprintf(stderr, "\t\t validCall: findCallee returned false; continuing\n");
                    continue;
                } 
                if (!targetFunc) {
                    // There was unresolved control flow
                    addToCache(callInsn, calleeFrameFuncs);
                    return true;
                }

                // Is there overlap between target of edge and callee frame?
                if (isMatch(targetFunc, calleeFrameFuncs)) {
                    if (DEBUG) fprintf(stderr, "\t\t vaildCall: found valid call edge\n");
                    addToCache(callInsn, calleeFrameFuncs);
                    return true;
                } else {
                    if (DEBUG) fprintf(stderr, "\t\t validCall: targetFunc %s not found in calleeFrameFuncs = {", targetFunc->name().c_str());
                    std::set<ParseAPI::Function *>::iterator cfIter;
                    for (cfIter = calleeFrameFuncs.begin(); cfIter != calleeFrameFuncs.end(); ++cfIter) {
                        if (DEBUG) fprintf(stderr, " %s,", (*cfIter)->name().c_str());
                    }
                    if (DEBUG) fprintf(stderr, "}\n");

                    // try tail-call detection
                    if (validTailCall(walker, targetFunc, calleeFrameFuncs)) {
                        if (DEBUG) fprintf(stderr, "\t\t validCall: found valid tail call path from %s->%s @ 0x%lx\n",
                                callerName.c_str(), 
                                targetFunc->name().c_str(), 
                                callInsn);
                        addToCache(callInsn, calleeFrameFuncs);
                        return true;
                    } 
                }
            }
        }
    }

    // If we couldn't find a srcblock in any of the code regions, try something else...

    // Check if we have a cached manaul instruction here
    std::tr1::unordered_map<Address, std::pair<InstructionAPI::Instruction::Ptr, ParseAPI::Function*> >::iterator findManual;
    findManual = cachedManualInstructions.find(relAddr);
    if (findManual == cachedManualInstructions.end()) {
        bool foundRelAddr = false;
        
        if (DEBUG) fprintf(stderr, "\t\t validCall: parsing shenanigans\n");
        std::map<ParseAPI::CodeObject*, std::map<Address, ParseAPI::Function*> >::iterator findPreproc = preproc.find(co);
        if (findPreproc == preproc.end()) {
            if (DEBUG) fprintf(stderr, "\t\t validCall: could not find preproc for this code object...\n");
            return false;
        }

        std::map<Address, ParseAPI::Function*> preproc = findPreproc->second;
        std::map<Address, ParseAPI::Function *>::const_iterator pre_iter = preproc.lower_bound(relAddr);
        if (pre_iter == preproc.begin()) {
            if (DEBUG) fprintf(stderr, "\t\t validCall: preproc.find failed\n");
            return false;
        }

        pre_iter--;

        // Decode from pre_iter->first to relAddr
        ParseAPI::Function * lastFunc = pre_iter->second;
        Address lastAddr = pre_iter->first;
        if (DEBUG) fprintf(stderr, "\t\t validCall: lastAddr = 0x%lx\n", lastAddr);
        Stackwalker::ProcessState * procState = walker->getProcessState();
        unsigned size = relAddr - lastAddr + 20;
        unsigned char buffer[size];
        if (!procState->readMem(buffer, lastAddr+libLoad, size)) {
            if (DEBUG) fprintf(stderr, "\t\t validCall: readMem(0x%lx, %d) failed\n", lastAddr+libLoad, size);
            return false;
        }

        InstructionAPI::Instruction::Ptr tmp;
        InstructionAPI::InstructionDecoder d(buffer, size, arch);

        tmp = d.decode();

        Address curAddr = lastAddr;

        while (tmp) {
            InstructionAPI::Instruction::Ptr curInsn = tmp;
            cachedManualInstructions.insert(make_pair(curAddr, make_pair(curInsn, lastFunc)));

            if (curAddr > relAddr) {
                break;
            }                

            if (curAddr == relAddr) {
                foundRelAddr = true;
                // continue parsing here to cache the entire gap + a few 
            }

            curAddr += tmp->size();
            tmp = d.decode();
        }

        if (!foundRelAddr) {
            return false;
        }
    }

    if (DEBUG) fprintf(stderr, "\t\t validCall: found an instruction at relAddr 0x%lx\n", relAddr);
    // Find call target, continue on our merry way...
    Address callTarget;
    manualFindCalleeAddress(insn, relAddr, calleeFrameFuncs, callTarget);
    if (callTarget == (Address)-1) {
        // There was unresolved control flow
        if (DEBUG) fprintf(stderr, "\t\t validCall: found unresolved control flow\n");
        addToCache(callInsn, calleeFrameFuncs);
        return true;
    } else {
        targetFunc = NULL;
        if (findCalleeFunc(walker, relAddr, cs, targetFunc, libLoad, callTarget)) {
            // Is there overlap between target of edge and callee frame?
            if (isMatch(targetFunc, calleeFrameFuncs)) {
                if (DEBUG) fprintf(stderr, "\t\t vaildCall: found valid call edge\n");
                addToCache(callInsn, calleeFrameFuncs);
                return true;
            } else {
                if (DEBUG) fprintf(stderr, "\t\t validCall: targetFunc %s not found in calleeFrameFuncs = {", targetFunc->name().c_str());
                std::set<ParseAPI::Function *>::iterator cfIter;
                for (cfIter = calleeFrameFuncs.begin(); cfIter != calleeFrameFuncs.end(); ++cfIter) {
                    if (DEBUG) fprintf(stderr, " %s,", (*cfIter)->name().c_str());
                }
                if (DEBUG) fprintf(stderr, "}\n");

                // try tail-call detection
                if (validTailCall(walker, targetFunc, calleeFrameFuncs)) {
                    if (DEBUG) fprintf(stderr, "\t\t validCall: found valid tail call path from %s->%s @ 0x%lx\n",
                            callerName.c_str(), 
                            targetFunc->name().c_str(), 
                            callInsn);
                    addToCache(callInsn, calleeFrameFuncs);
                    return true;
                } 
            }   
        }
    }

    if (DEBUG) fprintf(stderr, "\t\t validCall: could not find valid call edge from %s->%s @ 0x%lx\n", 
            callerName.c_str(), 
            calleeName.c_str(),
            callInsn);
    return false;
}


