/* 
 * ROPStop is Copyright (c) 2012-2014 Barton P. Miller. Written by Emily R. Jacobson.
 *
 * This file is part of ROPStop.
 *
 * ROPStop is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * ROPStop is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with ROPStop.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>

#include "dyntypes.h"

#include "PCProcess.h"
#include "CodeSource.h"
#include "walker.h"
#include "procstate.h"

#include "Symtab.h"
#include "SymtabReader.h"

#include "CodeObject.h"
#include "CodeSource.h"
#include "CFG.h"

#include "Instruction.h"
#include "InstructionDecoder.h"

#include "SymEval.h"
#include "slicing.h"

/* Local includes */
#include "validPC.h"
#include "globals.h"
#include "accessors.h"

using namespace std;
using namespace Dyninst;

std::tr1::unordered_map<Address, std::pair<InstructionAPI::Instruction::Ptr, ParseAPI::Block *> > validPCs;

bool validPC(Stackwalker::Walker * walker,
        PID pid, 
        MachRegisterVal pc, 
        ParseAPI::CodeObject * co, 
        std::set<ParseAPI::CodeRegion *> & regions,
        std::string libName,
        Address libLoad,
        unsigned libSize,
        InstructionAPI::Instruction::Ptr & instr,
        ParseAPI::Block *& curBlock)
{
    
    std::tr1::unordered_map<Address, std::pair<InstructionAPI::Instruction::Ptr, ParseAPI::Block *> >::iterator vIter;
    vIter = validPCs.find((Address)pc);
    if (vIter != validPCs.end()) {
        if (DEBUG) fprintf(stderr, "\t validPC: found cached result\n");
        std::pair<InstructionAPI::Instruction::Ptr, ParseAPI::Block *> result = vIter->second;
        instr = result.first;
        curBlock = result.second;
        return true;
    }

    /* Let's actually check that PC points to a valid instruction. */
    Address relAddr = getRelAddr(pc, libName, libLoad);
    if (DEBUG) fprintf(stderr, "\t validPC: check that PC (relAddr = 0x%lx) points to valid instruction\n",
            relAddr);


    std::set<ParseAPI::Function *> funcs;
    ParseAPI::CodeRegion * cr;
    std::set<ParseAPI::CodeRegion *>::iterator rIter;
    std::set<ParseAPI::Block *> blocks;
    ParseAPI::CodeObject::funclist::iterator fIter;
    ParseAPI::Function::blocklist::iterator blistIter;
    std::set<ParseAPI::Block *>::iterator bIter;
    ParseAPI::Block * block;
    void * bufferPtr;
    InstructionAPI::Instruction::Ptr tmp;
    Address loc;
    for (rIter = regions.begin(); rIter != regions.end(); ++rIter) {
        cr = *rIter; 

        if (DEBUG) fprintf(stderr, "\t\t validPC: Checking region 0x%lx-0x%lx\n", cr->low(), cr->high());

        if (!co->findBlocks(cr, relAddr, blocks)) {
            if (DEBUG) fprintf(stderr, "\t\t validPC: Could not find block that contains relAddr\n");
            if (co->findBlocks(cr, pc, blocks)) {
                if (DEBUG) fprintf(stderr, "\t\t validPC: Found block that contains PC\n");
            }

            if (!co->findFuncs(cr, relAddr, funcs)) {
                if (DEBUG) fprintf(stderr, "\t\t validPC: Could not find any func that contains relAddr\n");

                ParseAPI::CodeObject::funclist allFuncs = co->funcs();
                for (fIter = allFuncs.begin(); fIter != allFuncs.end(); ++fIter) {

                    ParseAPI::Function::blocklist allBlocks = (*fIter)->blocks();
                }
            }

        } else {
            /* Iterate through block to check for instruction */
            for (bIter = blocks.begin(); bIter != blocks.end(); ++bIter) {
                block = *bIter;

                if (DEBUG) fprintf(stderr, "\t\t validPC: Checking block at 0x%lx\n", block->start());

                bufferPtr = cr->getPtrToInstruction(block->start());
                if (!bufferPtr) {
                    if (DEBUG) fprintf(stderr, "\t\t validPC: Couldn't get ptr to instr at block start\n");
                } else {
                    InstructionAPI::InstructionDecoder d(bufferPtr, block->size(), walker->getProcessState()->getArchitecture());
                    // Decode all instructions in the buffer
                    tmp = d.decode();
                    loc = block->start();
                    while (tmp) {
                        if (loc == relAddr) {
                            // If we've found the instruction pc points to, stop!
                            if (DEBUG) fprintf(stderr, "\t\t validPC: Found instr %s at PC\n", tmp->format().c_str());
                            instr = tmp;
                            curBlock = block; 
                            validPCs.insert(make_pair(pc, make_pair(instr, curBlock)));
                            return true;
                        } else {
                            // Else, decode the next instruction
                            loc += tmp->size();
                            tmp = d.decode();
                        }
                    }
                } 
            }
        }

        if (DEBUG) fprintf(stderr, "\t\t COULD NOT FIND instr at PC\n");
        return false;
    }

    return false;
}
