/* 
 * ROPStop is Copyright (c) 2012-2014 Barton P. Miller. Written by Emily R. Jacobson.
 *
 * This file is part of ROPStop.
 *
 * ROPStop is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * ROPStop is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with ROPStop.  If not, see <http://www.gnu.org/licenses/>.
 */

/* Includes */
#include <stdio.h>
#include <getopt.h>
#include <iostream>
#include <sstream>

#include "dyntypes.h"

/* ProcControl includes */
#include "PCProcess.h"
#include "PCErrors.h"

/* Stackwalker includes */
#include "walker.h"
#include "frame.h"
#include "basetypes.h"
#include "framestepper.h"
#include "steppergroup.h"
#include "swk_errors.h"

/* InstructionAPI includes */
#include "Instruction.h"
#include "InstructionDecoder.h"

#include "globals.h"

using namespace std;
using namespace Dyninst;

std::tr1::unordered_map<Address, std::pair<InstructionAPI::Instruction::Ptr, Address> > validRAs;

bool getInsn(ProcControlAPI::Process::const_ptr proc,
        unsigned char * ptr, 
        int size,
        InstructionAPI::Instruction::Ptr & insn) 
{
    InstructionAPI::InstructionDecoder d(ptr, size, arch);
    insn = d.decode();
    
    if (insn) return true;
    else return false;

}

bool isInsnACall(InstructionAPI::Instruction::Ptr insn)
{
    entryID id = insn->getOperation().getID();
    return (id == e_call || id == e_syscall || id == e_sysenter);
}

void printBuffer(unsigned char * buffer, int size)
{
    for (int i = 0; i < size; i++) {
        fprintf(stderr, "buffer[%d] = %x\n", i, buffer[i]);
    }
}

bool validRA(Stackwalker::Walker * walker,
        ProcControlAPI::Process::const_ptr proc, 
        Address addr,
        InstructionAPI::Instruction::Ptr & insn,
        Address & insn_addr)
{

    // Do we have a cached copy?
    std::tr1::unordered_map<Address,std::pair<InstructionAPI::Instruction::Ptr,Address> >::iterator vIter;
    vIter = validRAs.find(addr);
    if (vIter != validRAs.end()) {
        if (DEBUG) fprintf(stderr, "\t\t validRA: found cached result\n");
        std::pair<InstructionAPI::Instruction::Ptr, Address> result = vIter->second;
        insn = result.first;
        insn_addr = result.second;
        return true;
    }

    bool found = false;
    int max_instr_size = 8;

    if (DEBUG) fprintf(stderr, "\t\t validRA: checking address 0x%lx for prev instr is call\n", addr);

    Stackwalker::ProcessState * procState = walker->getProcessState();
    Address call;

    unsigned char buffer[max_instr_size];
    for (unsigned size = max_instr_size; size > 0; size--) {
        call = addr - size;
        if (procState->readMem(buffer, call, size)) break;
    }

    unsigned char * bufferPtr = buffer;

    InstructionAPI::Instruction::Ptr tmp;
    InstructionAPI::Instruction::Ptr prevInsn;
    unsigned int aligned;
    int insn_size;
    
    for (unsigned size = addr - call; size > 0; size--) {

        InstructionAPI::InstructionDecoder d(bufferPtr, size, arch);

        tmp = d.decode();
        prevInsn = tmp;

        aligned = 0;

        // Decode all instructions in the buffer
        while (tmp) {
            if (tmp->size() > size) break;

            aligned += tmp->size();
            prevInsn = tmp;
            tmp = d.decode();
        }

        // prevInsn was the last valid instruction we found
        // is it (a) aligned and (b) a call?
        if ( (aligned == size) && (isInsnACall(prevInsn)) ) {
            insn = prevInsn;
            insn_size = insn->size();
            insn_addr = addr - insn_size;
            if (DEBUG) fprintf(stderr, "\t\t validRA: found valid prev CALL instruction [%s] at 0x%lx\n", insn->format().c_str(), insn_addr);
            found = true;

        } 
        if (DEBUG && tmp) fprintf(stderr, "\t\t validRA: found instr %s\n", tmp->format().c_str());
        bufferPtr++;
        continue;
    }

    if (found) {
        validRAs.insert(make_pair(addr, make_pair(insn, insn_addr)));
        return true;
    }

    return false;
}


